/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.utils;

import static org.fortiss.tooling.kernel.utils.EcoreUtils.getEClassForClass;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getParentsWithType;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickFirstInstanceOf;

import java.util.List;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IDerivedAnnotation;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.service.IPersistencyService;

// TODO: Some of the methods in this class also apply to {@link IModelElementSpecification}s, not only
// {@link IAnnotatedSpecification}s.
/**
 * Utility methods for dealing with {@link IAnnotatedSpecification}s.
 * 
 * @author diewald
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating RED Hash:
 */
public class AnnotationUtils {

	// TODO This function is a workaround which might no longer be required. Before the IDE dev
	// tools update on 2014/12/01, the EMF code generator failed when generic abstract base
	// classes are referenced from another ecore file.

	/**
	 * Wrapper for returning derived values of {@link IDerivedAnnotation}s. If the user input is
	 * preferred by the concrete annotation, this function returns the user input if any is set.
	 * Otherwise it determines the calculated value, if the value can be calculated. If the value
	 * cannot be calculated, and a user input is given which is not preferred it is returned
	 * instead. If none of the previously mentioned cases applies, null is returned.
	 * 
	 * @param annotation
	 *            The annotation to which this wrapper shall be applied to.
	 * @return See above.
	 */
	public static <T> T getValue(IDerivedAnnotation<T> annotation) {
		if(annotation.getUserAnnotatedValue() != null &&
				(annotation.isUserAnnotatedValuePreferred() || annotation.getDerivedValue() == null)) {
			return annotation.getUserAnnotatedValue();
		}
		return annotation.getDerivedValue();
	}

	/**
	 * Returns the annotation of a given {@link EClass} from a given {@code modelElement}. If the
	 * annotation does not exist yet, it is created and added to the specification list of the
	 * {@code modelElement}.
	 */
	@SuppressWarnings("unchecked")
	private static <T extends IAnnotatedSpecification> T getAnnotation(
			final IModelElement modelElement, EClass eClass) {

		final T annotation =
				(T)pickFirstInstanceOf((Class<IAnnotatedSpecification>)eClass.getInstanceClass(),
						modelElement.getSpecifications());

		if(annotation != null) {
			return annotation;
		}

		final T newAnnotation = (T)EcoreUtil.create(eClass);
		ITopLevelElement modelContext =
				IPersistencyService.INSTANCE.getTopLevelElementFor(modelElement);
		modelContext.runAsCommand(new Runnable() {

			@Override
			public void run() {
				modelElement.getSpecifications().add(newAnnotation);
			}
		});

		return newAnnotation;
	}

	/**
	 * Returns the annotation of a given {@code Class<T>} from a given {@code modelElement}. If the
	 * annotation does not exist yet, it is created and added to the specification list of the
	 * {@code modelElement}.
	 */
	public static <T extends IAnnotatedSpecification> T getAnnotation(IModelElement modelElement,
			Class<T> clazz) {

		EClass eClass = getEClassForClass(clazz);

		if(eClass != null) {
			// Retrieve (potentially create) annotation
			return getAnnotation(modelElement, eClass);
		}

		return null;
	}

	/**
	 * Returns the {@link IAnnotatedSpecification}s matching the type 'T' of
	 * the parents of the given {@link IModelElement}.
	 * 
	 * @param startElement
	 *            The {@link IModelElement} to start the search from.
	 * @param clazz
	 *            The {@link IAnnotatedSpecification}s to be searched for.
	 * @return {@link EList} of found annotations.
	 */
	public static <T extends IAnnotatedSpecification> EList<T> getParentAnnotationsWithType(
			IModelElement startElement, Class<T> clazz) {
		EList<T> parentSpecificationList = new BasicEList<T>();
		List<IModelElement> parentsElements = getParentsWithType(startElement, IModelElement.class);
		for(IModelElement currentParentElement : parentsElements) {
			parentSpecificationList.add(pickFirstInstanceOf(clazz,
					currentParentElement.getSpecifications()));
		}
		return parentSpecificationList;
	}
}
